# TRIDENT

__title__ = 'Trident'
__description__ = 'Aquarium scripting'
__url__ = 'https://klavinslab.org/trident'
__version__ = '0.0.22'
__author__ = 'Justin Vrana, Eric Klavins, Ben Keller'
__author_email__ = "jvrana@uw.edu"
